//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/compat.rs: Compatibility code for different libcs
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    ffi::{OsStr, OsString},
    os::{fd::RawFd, unix::ffi::OsStrExt},
    path::{Path, PathBuf},
};

use nix::errno::Errno;
use nonempty::NonEmpty;

/// This structure represents the Linux 32-bit data structure 'struct iovec'
#[repr(C)]
pub struct iovec32 {
    iov_base: u32,
    iov_len: u32,
}

impl From<iovec32> for nix::libc::iovec {
    fn from(src: iovec32) -> Self {
        nix::libc::iovec {
            iov_base: src.iov_base as *mut _,
            iov_len: src.iov_len as usize,
        }
    }
}

/// This structure represents the Linux 32-bit data structure 'struct mmmsghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct mmsghdr32 {
    pub msg_hdr: msghdr32,
    pub msg_len: u32,
}

/// This structure represents the Linux native data structure 'struct mmsghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct mmsghdr {
    pub msg_hdr: msghdr,
    pub msg_len: nix::libc::c_uint,
}

impl From<mmsghdr32> for mmsghdr {
    fn from(src: mmsghdr32) -> Self {
        mmsghdr {
            msg_hdr: msghdr::from(src.msg_hdr),
            msg_len: src.msg_len,
        }
    }
}

impl From<mmsghdr> for mmsghdr32 {
    fn from(src: mmsghdr) -> Self {
        mmsghdr32 {
            msg_hdr: msghdr32::from(src.msg_hdr),
            msg_len: src.msg_len,
        }
    }
}

/// This structure represents the Linux 32-bit data structure 'struct cmsghdr'
#[repr(C)]
pub struct cmsghdr32 {
    pub cmsg_len: u32,
    pub cmsg_level: i32,
    pub cmsg_type: i32,
}

/// This structure represents the Linux native data structure 'struct cmsghdr'
#[repr(C)]
pub struct cmsghdr {
    pub cmsg_len: nix::libc::size_t,
    pub cmsg_level: nix::libc::c_int,
    pub cmsg_type: nix::libc::c_int,
}

impl From<cmsghdr32> for cmsghdr {
    fn from(src: cmsghdr32) -> Self {
        cmsghdr {
            cmsg_len: src.cmsg_len as nix::libc::size_t,
            cmsg_level: src.cmsg_level,
            cmsg_type: src.cmsg_type,
        }
    }
}

/// This structure represents the Linux 32-bit data structure 'struct msghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct msghdr32 {
    // In 32-bit systems, pointers are 32-bit.
    pub msg_name: u32,       // Use u32 to represent a 32-bit pointer.
    pub msg_namelen: u32,    // socklen_t is typically 32-bit.
    pub msg_iov: u32,        // Use u32 to represent a 32-bit pointer to iovec.
    pub msg_iovlen: u32,     // size_t is 32-bit on 32-bit systems.
    pub msg_control: u32,    // Use u32 to represent a 32-bit pointer.
    pub msg_controllen: u32, // size_t is 32-bit on 32-bit systems.
    pub msg_flags: i32,      // c_int remains the same (32-bit).
}

/// This structure represents the Linux native data structure 'struct msghdr'
#[derive(Copy, Clone)]
#[repr(C)]
pub struct msghdr {
    pub msg_name: *mut nix::libc::c_void,
    pub msg_namelen: nix::libc::socklen_t,
    pub msg_iov: *mut nix::libc::iovec,
    pub msg_iovlen: nix::libc::size_t,
    pub msg_control: *mut nix::libc::c_void,
    pub msg_controllen: nix::libc::size_t,
    pub msg_flags: nix::libc::c_int,
}

impl From<msghdr32> for msghdr {
    fn from(msg: msghdr32) -> Self {
        msghdr {
            msg_name: msg.msg_name as *mut nix::libc::c_void,
            msg_namelen: msg.msg_namelen as nix::libc::socklen_t,
            msg_iov: msg.msg_iov as *mut nix::libc::iovec,
            msg_iovlen: msg.msg_iovlen as nix::libc::size_t,
            msg_control: msg.msg_control as *mut nix::libc::c_void,
            msg_controllen: msg.msg_controllen as nix::libc::size_t,
            msg_flags: msg.msg_flags as nix::libc::c_int,
        }
    }
}

#[allow(clippy::unnecessary_cast)]
#[allow(clippy::cast_possible_truncation)]
impl From<msghdr> for msghdr32 {
    fn from(msg: msghdr) -> Self {
        msghdr32 {
            msg_name: msg.msg_name as u32,
            msg_namelen: msg.msg_namelen as u32,
            msg_iov: msg.msg_iov as u32,
            msg_iovlen: msg.msg_iovlen as u32,
            msg_control: msg.msg_control as u32,
            msg_controllen: msg.msg_controllen as u32,
            msg_flags: msg.msg_flags as i32,
        }
    }
}

/// This structure represents the Linux data structure `struct statx_timestamp`
#[repr(C)]
pub(crate) struct statx_timestamp {
    tv_sec: i64,
    tv_nsec: u32,
    __statx_timestamp_pad1: [i32; 1],
}

/// This structure represents the Linux data structure `struct statx`
#[repr(C)]
pub(crate) struct statx {
    stx_mask: u32,       // What results were written [uncond]
    stx_blksize: u32,    // Preferred general I/O size [uncond]
    stx_attributes: u64, // Flags conveying information about the file [uncond]

    stx_nlink: u32, // Number of hard links
    stx_uid: u32,   // User ID of owner
    stx_gid: u32,   // Group ID of owner
    stx_mode: u16,  // File mode
    __statx_pad1: [u16; 1],

    stx_ino: u64,             // Inode number
    stx_size: u64,            // File size
    stx_blocks: u64,          // Number of 512-byte blocks allocated
    stx_attributes_mask: u64, // Mask to show what's supported in stx_attributes

    stx_atime: statx_timestamp, // Last access time
    stx_btime: statx_timestamp, // File creation time
    stx_ctime: statx_timestamp, // Last attribute change time
    stx_mtime: statx_timestamp, // Last data modification time

    stx_rdev_major: u32, // Device ID of special file [if bdev/cdev]
    stx_rdev_minor: u32,
    stx_dev_major: u32, // ID of device containing file [uncond]
    stx_dev_minor: u32,

    stx_mnt_id: u64,
    stx_dio_mem_align: u32,    // Memory buffer alignment for direct I/O
    stx_dio_offset_align: u32, // File offset alignment for direct I/O

    __statx_pad2: [u64; 12], // Spare space for future expansion
}

/// This struct represents a directory entry.
#[derive(Clone)]
pub(crate) struct DirEntry {
    /// The `dirent64` structure.
    pub(crate) dirent: Box<[u8]>,

    /// Size of the file name, in bytes.
    pub(crate) namelen: usize,
}

impl std::fmt::Debug for DirEntry {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_tuple("DirEntry").field(&self.name()).finish()
    }
}

impl DirEntry {
    /// Return a `PathBuf` representing the name of the directory entry.
    pub fn to_path_buf(&self) -> PathBuf {
        Path::new(&self.name()).to_path_buf()
    }

    /// Return the name of the directory entry.
    pub fn name(&self) -> OsString {
        // SAFETY: We trust self.dirent points to a valid `dirent64` structure.
        OsStr::from_bytes(unsafe { self.name_bytes() }).to_os_string()
    }

    unsafe fn name_bytes(&self) -> &[u8] {
        let dirent: *const nix::libc::dirent64 = Box::as_ref(&self.dirent).as_ptr().cast();
        let d_name: *const nix::libc::c_char = (*dirent).d_name.as_ptr();
        std::slice::from_raw_parts(d_name.cast(), self.namelen)
    }
}

/// Represents a list of directory entries as returned by `getdents64` system call.
pub(crate) type Directory = NonEmpty<DirEntry>;

/// Retrieve directory entries from an open directory file descriptor.
///
/// # Parameters
///
/// - `fd`: The open directory file descriptor.
///
/// # Returns
///
/// `Some(Directory)` containing the directory entries,
/// `None` when there are no more entries.
///
/// # Safety
///
/// This function calls the `getdents64` system call directly which is an unsafe function.
/// Ensure that `fd` is a valid open directory file descriptor to avoid undefined behavior.
pub(crate) fn getdents(fd: RawFd, count: usize) -> Result<Option<Directory>, Errno> {
    let mut buffer = vec![0u8; count];

    let size = loop {
        // SAFETY: In libc, we trust.
        match unsafe { getdents64(fd, buffer.as_mut_ptr().cast(), buffer.len()) } {
            0 => {
                return Ok(None);
            }
            error if error < 0 => {
                #[allow(clippy::arithmetic_side_effects)]
                #[allow(clippy::cast_possible_truncation)]
                let error = -error as i32;
                if error == Errno::EINTR as i32 {
                    continue;
                }
                return Err(Errno::from_i32(error));
            }
            size =>
            {
                #[allow(clippy::cast_possible_truncation)]
                #[allow(clippy::cast_sign_loss)]
                break size as usize
            }
        }
    };

    // Parsing the buffer to extract DirEntry structures
    let mut entries = Vec::new();
    let mut offset = 0;
    while offset < size {
        // SAFETY: In libc, we trust.
        unsafe {
            let dirent: *const nix::libc::dirent64 = buffer.as_ptr().add(offset).cast();
            let d_reclen = (*dirent).d_reclen as usize;

            // Copy the bytes of the `dirent64` structure from the buffer to a
            // memory owned by `DirEntry`.
            let entry = std::slice::from_raw_parts(dirent.cast(), d_reclen)
                .to_owned()
                .into_boxed_slice();

            entries.push(DirEntry {
                dirent: entry,
                namelen: nix::libc::strlen((*dirent).d_name.as_ptr()),
            });

            offset = offset.saturating_add(d_reclen);
        }
    }

    if entries.is_empty() {
        Ok(None)
    } else {
        #[allow(clippy::disallowed_methods)]
        Ok(Some(NonEmpty::from_vec(entries).unwrap()))
    }
}

/// Wrapper for the `getdents64` syscall
#[inline(always)]
unsafe fn getdents64(
    fd: nix::libc::c_int,
    buf: *mut nix::libc::c_void,
    bytes: nix::libc::size_t,
) -> nix::libc::c_long {
    nix::libc::syscall(nix::libc::SYS_getdents64, fd, buf, bytes)
}
