//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/hash.rs: Utilities for hashing
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    collections::{hash_map::DefaultHasher, HashSet},
    hash::{BuildHasher, Hash, Hasher},
    io::BufRead,
};

use nix::errno::Errno;

/// Avoid duplicate hashing while using `HashSet` with u64 keys.
pub struct NoHasher {
    value: u64,
}

impl Hasher for NoHasher {
    fn write(&mut self, _bytes: &[u8]) {
        unreachable!("NoHasher should only be used for u64 keys");
    }

    fn write_u64(&mut self, i: u64) {
        self.value = i;
    }

    fn finish(&self) -> u64 {
        self.value
    }
}

/// A builder for creating instances of `NoHasher`.
#[derive(Clone)]
pub struct NoHasherBuilder;

impl Default for NoHasherBuilder {
    fn default() -> Self {
        Self
    }
}

impl BuildHasher for NoHasherBuilder {
    type Hasher = NoHasher;

    fn build_hasher(&self) -> Self::Hasher {
        NoHasher { value: 0 }
    }
}

/// A `HashSet` with no hashers.
pub type NoHashSet = HashSet<u64, NoHasherBuilder>;

/// Hash a string slice.
pub fn hash_str(s: &str) -> u64 {
    let mut hasher = DefaultHasher::new();
    s.hash(&mut hasher);
    hasher.finish()
}

/// Defines hash functions supported by syd.
#[derive(Debug, Clone, Copy)]
pub enum HashFunc {
    /// Md5
    Md5,
    /// SHA-1.
    Sha1,
    /// SHA3-256
    Sha256,
    /// SHA3-384
    Sha384,
    /// SHA3-512
    Sha512,
}

/// Calculate sha{1,256,512} of the given buffered reader.
/// Returns a byte array.
pub fn hash<R: BufRead>(mut reader: R, func: HashFunc) -> Result<Vec<u8>, Errno> {
    match func {
        HashFunc::Md5 => {
            let mut context = md5::Context::new();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    context.consume(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            let digest = context.compute();
            let dbytes: [u8; 16] = digest.into();
            Ok(dbytes.to_vec())
        }
        HashFunc::Sha1 => {
            let mut hasher = rs_sha1::Sha1Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha1::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
        HashFunc::Sha256 => {
            let mut hasher = rs_sha3_256::Sha3_256Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha3_256::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
        HashFunc::Sha384 => {
            let mut hasher = rs_sha3_384::Sha3_384Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha3_384::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
        HashFunc::Sha512 => {
            let mut hasher = rs_sha3_512::Sha3_512Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_i32(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha3_512::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
    }
}
