//
// syd: seccomp and landlock based application sandbox with support for namespaces
// src/test/test.rs: Integration tests
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(clippy::disallowed_methods)]

use std::{
    env,
    fs::{metadata, remove_file, set_permissions, File},
    io::{BufReader, Read, Write},
    os::{
        fd::{AsRawFd, FromRawFd},
        unix::{fs::PermissionsExt, process::ExitStatusExt},
    },
    process::{Command, Stdio},
};

use nix::{
    sys::stat::{umask, Mode},
    unistd::{unlink, Pid},
};
use syd::hash::HashFunc;

use crate::{assert, assert_eq, fixup, ignore, util::*};

const EX_SIGIOT: i32 = 128 + nix::libc::SIGIOT;
const EX_SIGSEGV: i32 = 128 + nix::libc::SIGSEGV;

/// Represents a test case.
pub type Test<'a> = (&'a str, fn() -> TestResult);

macro_rules! test_entry {
    ($func:expr) => {
        (stringify!($func), $func)
    };
}

/// List of integration tests.
pub const TESTS: &[Test] = &[
    test_entry!(test_syd_version),
    test_entry!(test_syd_export),
    test_entry!(test_syd_true_returns_success),
    test_entry!(test_syd_true_returns_success_with_many_processes),
    test_entry!(test_syd_true_returns_success_with_many_threads),
    test_entry!(test_syd_false_returns_failure),
    test_entry!(test_syd_true_returns_failure_with_many_processes),
    test_entry!(test_syd_true_returns_failure_with_many_threads),
    test_entry!(test_syd_sigint_returns_130),
    test_entry!(test_syd_sigabrt_returns_134),
    test_entry!(test_syd_sigkill_returns_137),
    test_entry!(test_syd_reap_zombies_bare),
    test_entry!(test_syd_reap_zombies_wrap),
    test_entry!(test_syd_whoami_returns_root_fake),
    test_entry!(test_syd_whoami_returns_root_user),
    test_entry!(test_syd_environment_filter),
    test_entry!(test_syd_environment_harden),
    test_entry!(test_syd_lock),
    test_entry!(test_syd_lock_exec),
    test_entry!(test_syd_chk),
    test_entry!(test_syd_shellexpand_01),
    test_entry!(test_syd_shellexpand_02),
    test_entry!(test_syd_shellexpand_03),
    test_entry!(test_syd_shellexpand_04),
    test_entry!(test_syd_lock_personality),
    test_entry!(test_syd_mdwe_protections),
    test_entry!(test_syd_mdwe_relax),
    test_entry!(test_syd_mmap_fixed_null),
    test_entry!(test_syd_load_library),
    test_entry!(test_syd_load_library_noexec),
    test_entry!(test_syd_load_library_abort_after_load),
    test_entry!(test_syd_load_library_abort_at_startup),
    test_entry!(test_syd_read_sandbox_open_allow),
    test_entry!(test_syd_read_sandbox_open_deny),
    test_entry!(test_syd_stat_sandbox_chdir_allow),
    test_entry!(test_syd_stat_sandbox_chdir_hide),
    test_entry!(test_syd_stat_sandbox_stat_allow),
    test_entry!(test_syd_stat_sandbox_stat_hide),
    test_entry!(test_syd_stat_sandbox_getdents_allow),
    test_entry!(test_syd_stat_sandbox_getdents_hide),
    test_entry!(test_syd_stat_bypass_with_read),
    test_entry!(test_syd_stat_bypass_with_write),
    test_entry!(test_syd_stat_bypass_with_exec),
    test_entry!(test_syd_write_sandbox_open_allow),
    test_entry!(test_syd_write_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_open_allow),
    test_entry!(test_syd_exec_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_deny_shebang),
    test_entry!(test_syd_network_sandbox_connect_ipv4_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv4_deny),
    test_entry!(test_syd_network_sandbox_connect_ipv6_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv6_deny),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv4_failure),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv4_success),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv6_failure),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv6_success),
    test_entry!(test_syd_handle_toolong_unix_connect),
    test_entry!(test_syd_handle_toolong_unix_sendto),
    test_entry!(test_syd_handle_toolong_unix_sendmsg),
    test_entry!(test_syd_sendmsg_scm_rights_one),
    test_entry!(test_syd_sendmsg_scm_rights_many),
    test_entry!(test_syd_sendmmsg),
    test_entry!(test_syd_exit_wait_all),
    test_entry!(test_syd_exit_wait_pid),
    test_entry!(test_syd_cli_args_override_user_profile),
    test_entry!(test_syd_force_sandbox),
    test_entry!(test_syd_exp_symlink_toctou),
    test_entry!(test_syd_exp_symlinkat_toctou),
    test_entry!(test_syd_exp_ptrmod_toctou_chdir),
    test_entry!(test_syd_exp_ptrmod_toctou_exec_fail),
    test_entry!(test_syd_exp_ptrmod_toctou_exec_success),
    test_entry!(test_syd_exp_ptrmod_toctou_open),
    test_entry!(test_syd_exp_ptrmod_toctou_creat),
    test_entry!(test_syd_exp_ptrmod_toctou_opath),
    test_entry!(test_syd_io_uring_escape_strict),
    test_entry!(test_syd_io_uring_escape_unsafe),
    test_entry!(test_syd_opath_escape),
    test_entry!(test_syd_devfd_escape_chdir),
    test_entry!(test_syd_devfd_escape_chdir_relpath_1),
    test_entry!(test_syd_devfd_escape_chdir_relpath_2),
    test_entry!(test_syd_devfd_escape_chdir_relpath_3),
    test_entry!(test_syd_devfd_escape_chdir_relpath_4),
    test_entry!(test_syd_devfd_escape_chdir_relpath_5),
    test_entry!(test_syd_devfd_escape_chdir_relpath_6),
    test_entry!(test_syd_devfd_escape_chdir_relpath_7),
    test_entry!(test_syd_devfd_escape_chdir_relpath_8),
    test_entry!(test_syd_devfd_escape_chdir_relpath_9),
    test_entry!(test_syd_devfd_escape_chdir_relpath_10),
    test_entry!(test_syd_devfd_escape_chdir_relpath_11),
    test_entry!(test_syd_devfd_escape_chdir_relpath_12),
    test_entry!(test_syd_devfd_escape_chdir_relpath_13),
    test_entry!(test_syd_devfd_escape_chdir_relpath_14),
    test_entry!(test_syd_devfd_escape_chdir_relpath_15),
    test_entry!(test_syd_devfd_escape_chdir_relpath_16),
    test_entry!(test_syd_devfd_escape_chdir_relpath_17),
    test_entry!(test_syd_devfd_escape_chdir_relpath_18),
    test_entry!(test_syd_devfd_escape_chdir_relpath_19),
    test_entry!(test_syd_devfd_escape_chdir_relpath_20),
    test_entry!(test_syd_devfd_escape_open),
    test_entry!(test_syd_devfd_escape_open_relpath_1),
    test_entry!(test_syd_devfd_escape_open_relpath_2),
    test_entry!(test_syd_devfd_escape_open_relpath_3),
    test_entry!(test_syd_devfd_escape_open_relpath_4),
    test_entry!(test_syd_devfd_escape_open_relpath_5),
    test_entry!(test_syd_devfd_escape_open_relpath_6),
    test_entry!(test_syd_devfd_escape_open_relpath_7),
    test_entry!(test_syd_devfd_escape_open_relpath_8),
    test_entry!(test_syd_devfd_escape_open_relpath_9),
    test_entry!(test_syd_devfd_escape_open_relpath_10),
    test_entry!(test_syd_devfd_escape_open_relpath_11),
    test_entry!(test_syd_devfd_escape_open_relpath_12),
    test_entry!(test_syd_devfd_escape_open_relpath_13),
    test_entry!(test_syd_devfd_escape_open_relpath_14),
    test_entry!(test_syd_devfd_escape_open_relpath_15),
    test_entry!(test_syd_devfd_escape_open_relpath_16),
    test_entry!(test_syd_devfd_escape_open_relpath_17),
    test_entry!(test_syd_devfd_escape_open_relpath_18),
    test_entry!(test_syd_devfd_escape_open_relpath_19),
    test_entry!(test_syd_devfd_escape_open_relpath_20),
    test_entry!(test_syd_procself_escape_chdir),
    test_entry!(test_syd_procself_escape_chdir_relpath_1),
    test_entry!(test_syd_procself_escape_chdir_relpath_2),
    test_entry!(test_syd_procself_escape_chdir_relpath_3),
    test_entry!(test_syd_procself_escape_chdir_relpath_4),
    test_entry!(test_syd_procself_escape_chdir_relpath_5),
    test_entry!(test_syd_procself_escape_chdir_relpath_6),
    test_entry!(test_syd_procself_escape_chdir_relpath_7),
    test_entry!(test_syd_procself_escape_open),
    test_entry!(test_syd_procself_escape_open_relpath_1),
    test_entry!(test_syd_procself_escape_open_relpath_2),
    test_entry!(test_syd_procself_escape_open_relpath_3),
    test_entry!(test_syd_procself_escape_open_relpath_4),
    test_entry!(test_syd_procself_escape_open_relpath_5),
    test_entry!(test_syd_procself_escape_open_relpath_6),
    test_entry!(test_syd_procself_escape_open_relpath_7),
    test_entry!(test_syd_procself_escape_relpath),
    test_entry!(test_syd_procself_escape_symlink),
    test_entry!(test_syd_procself_escape_symlink_within_container),
    test_entry!(test_syd_umask_bypass_077),
    test_entry!(test_syd_umask_bypass_277),
    test_entry!(test_syd_emulate_opath),
    test_entry!(test_syd_emulate_otmpfile),
    test_entry!(test_syd_honor_umask),
    test_entry!(test_syd_open_utf8_invalid),
    test_entry!(test_syd_exec_in_inaccessible_directory),
    test_entry!(test_syd_fstat_on_pipe),
    test_entry!(test_syd_fstat_on_socket),
    test_entry!(test_syd_fstat_on_deleted_file),
    test_entry!(test_syd_fstat_on_temp_file),
    test_entry!(test_syd_fchmodat_on_proc_fd),
    test_entry!(test_syd_linkat_on_fd),
    test_entry!(test_syd_block_ioctl_tiocsti),
    test_entry!(test_syd_block_prctl_ptrace),
    test_entry!(test_syd_block_dev_random),
    test_entry!(test_syd_block_dev_urandom),
    test_entry!(test_syd_kill_during_syscall),
    test_entry!(test_syd_open_toolong_path),
    test_entry!(test_syd_open_null_path),
    test_entry!(test_syd_utimensat_null),
    test_entry!(test_syd_normalize_path),
    test_entry!(test_syd_path_resolution),
    test_entry!(test_syd_remove_empty_path),
    test_entry!(test_syd_symlink_readonly_path),
    test_entry!(test_syd_open_trailing_slash),
    test_entry!(test_syd_openat_trailing_slash),
    test_entry!(test_syd_lstat_trailing_slash),
    test_entry!(test_syd_fstatat_trailing_slash),
    test_entry!(test_syd_mkdir_trailing_dot),
    test_entry!(test_syd_mkdirat_trailing_dot),
    test_entry!(test_syd_rmdir_trailing_slashdot),
    test_entry!(test_syd_mkdir_eexist_escape),
    test_entry!(test_syd_mkdirat_eexist_escape),
    test_entry!(test_syd_mknod_eexist_escape),
    test_entry!(test_syd_mknodat_eexist_escape),
    test_entry!(test_syd_fopen_supports_mode_e),
    test_entry!(test_syd_fopen_supports_mode_x),
    test_entry!(test_syd_link_no_symlink_deref),
    test_entry!(test_syd_link_posix),
    test_entry!(test_syd_linkat_posix),
    test_entry!(test_syd_getcwd_long),
    test_entry!(test_syd_creat_thru_dangling),
    test_entry!(test_syd_mkdirat_non_dir_fd),
    test_entry!(test_syd_blocking_udp4),
    test_entry!(test_syd_blocking_udp6),
    test_entry!(test_syd_close_on_exec),
    test_entry!(test_syd_exp_open_exclusive_restart),
    test_entry!(test_syd_open_exclusive_repeat),
    test_entry!(test_syd_pty_io_rust),
    test_entry!(test_syd_pty_io_gawk),
    test_entry!(test_syd_diff_dev_fd),
    test_entry!(test_syd_fifo_multiple_readers),
    test_entry!(test_syd_bind_unix_socket),
    test_entry!(test_syd_signal_protection),
    test_entry!(test_syd_exp_emulate_open_fifo_1),
    test_entry!(test_syd_exp_emulate_open_fifo_2),
    test_entry!(test_syd_deny_magiclinks),
    test_entry!(test_syd_lstat_magiclinks),
    test_entry!(test_syd_access_unsafe_paths),
    test_entry!(test_syd_unshare_user_bypass_limit),
    test_entry!(test_syd_exp_interrupt_mkdir),
    test_entry!(test_syd_exp_interrupt_bind_ipv4),
    test_entry!(test_syd_exp_interrupt_bind_unix),
    test_entry!(test_syd_exp_interrupt_connect_ipv4),
    //FIXME: This test should be done better.
    //test_entry!(test_syd_repetitive_clone),
    test_entry!(test_syd_exp_syscall_fuzz_bare),
    test_entry!(test_syd_exp_syscall_fuzz_wrap),
    test_entry!(test_syd_pid_fork_deny),
    test_entry!(test_syd_pid_thread_deny),
    test_entry!(test_syd_pid_fork_bomb),
    test_entry!(test_syd_pid_fork_bomb_asm),
    test_entry!(test_syd_pid_thread_bomb),
    test_entry!(test_syd_exp_pid_stress_ng_kill),
    test_entry!(test_syd_exp_pid_stress_ng_allow),
    test_entry!(test_syd_exp_pid_stress_ng_fork),
    test_entry!(test_syd_mem_alloc),
    test_entry!(test_syd_exp_mem_stress_ng_malloc_1),
    test_entry!(test_syd_exp_mem_stress_ng_malloc_2),
    test_entry!(test_syd_exp_mem_stress_ng_mmap),
];

// Tests if syd -V and --version works.
fn test_syd_version() -> TestResult {
    let status = syd().arg("-V").status().expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .stdout(Stdio::null())
        .stderr(Stdio::null())
        .arg("--version")
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if syd -E works.
fn test_syd_export() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .arg("-Ebpf")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-Epfc")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-EBPF")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-EPFC")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-EbPf")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-EPfc")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-Eb")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    let status = syd()
        .arg("-Ef")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Tests if `true` returns success under sandbox.
fn test_syd_true_returns_success() -> TestResult {
    let status = syd()
        .env("SYD_DO", "exit")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if `syd` returns success for a sandbox running many processes,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_processes() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fork")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "0", "8"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if `syd` returns success for a sandbox running many threads,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_threads() -> TestResult {
    let status = syd()
        .env("SYD_DO", "thread")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "0", "8"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if `false` returns failure under sandbox.
fn test_syd_false_returns_failure() -> TestResult {
    let status = syd()
        .arg("-mallow/exec,read,stat+/***")
        .arg("--")
        .arg("false")
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    let status = syd()
        .args(["-plib", "--", "false"])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many processes,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_processes() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fork")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "7", "8"])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 7, "status:{status:?}");
    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many threads,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_threads() -> TestResult {
    let status = syd()
        .env("SYD_DO", "thread")
        .arg("-mallow/exec,read,stat+/***")
        .args(["--", &SYD_DO, "7", "8"])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 7, "status:{status:?}");
    Ok(())
}

fn test_syd_sigint_returns_130() -> TestResult {
    if !is_program_available("kill") {
        eprintln!("Test requires kill, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["sh", "-c"])
        .arg(r#"exec kill -INT $$"#)
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 130);
    Ok(())
}

fn test_syd_sigabrt_returns_134() -> TestResult {
    if !is_program_available("kill") {
        eprintln!("Test requires kill, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["sh", "-c"])
        .arg(r#"exec kill -ABRT $$"#)
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 134);
    Ok(())
}

fn test_syd_sigkill_returns_137() -> TestResult {
    if !is_program_available("kill") {
        eprintln!("Test requires kill, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["sh", "-c"])
        .arg(r#"exec kill -KILL $$"#)
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 137);
    Ok(())
}

fn test_syd_reap_zombies_bare() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-ppaludis")
        .args(["bash", "-c"])
        .arg(
            r#"
set -e
for i in {1..10}; do
    ( sleep $i ) &
done
echo >&2 "Spawned 10 processes in the background."
echo >&2 "Disowning and exiting..."
disown
exit 42
        "#,
        )
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 42);

    Ok(())
}

fn test_syd_reap_zombies_wrap() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };
    let status = syd()
        .args(["-ppaludis", "-pcontainer"])
        .args(["bash", "-c"])
        .arg(
            r#"
set -e
for i in {1..10}; do
    ( sleep $i ) &
done
echo >&2 "Spawned 10 processes in the background."
echo >&2 "Disowning and exiting..."
disown
exit 42
        "#,
        )
        .status()
        .expect("execute syd");
    assert!(!status.success());
    assert_eq!(status.code().unwrap_or(127), 42);

    Ok(())
}

// Tests if `whoami` returns `root` with `root/fake:1`
fn test_syd_whoami_returns_root_fake() -> TestResult {
    let status = syd()
        .env("SYD_DO", "getuid")
        .arg("-plib")
        .args(["-mroot/fake:1", "--", &SYD_DO, "0"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if `whoami` returns `root` with `root/map:1`
fn test_syd_whoami_returns_root_user() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .env("SYD_DO", "getuid")
        .arg("-plib")
        .args(["-mroot/map:1", "--", &SYD_DO, "0"])
        .stderr(Stdio::inherit())
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Checks environment filtering
fn test_syd_environment_filter() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    const ENV: &str = "SAFE";
    env::set_var(ENV, "/var/empty");

    // Step 1: Allow by default
    let output = syd()
        .arg("-plib")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output1:{output}");

    // Step 2: Override with -evar=val
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}=/var/empty:/var/empty"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty:/var/empty", "output2:{output}");

    // Step 3: Unset with -evar
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output3:{output}");

    // Step 4: Pass-through with -evar=
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}="))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output4:{output}");

    env::remove_var(ENV);

    Ok(())
}

// Checks environment hardening and -e pass-through.
fn test_syd_environment_harden() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    const ENV: &str = "LD_LIBRARY_PATH";
    env::set_var(ENV, "/var/empty");

    // Step 1: Deny by default
    let output = syd()
        .arg("-plib")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output1:{output}");

    // Step 2: Override with -evar=val
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}=/var/empty:/var/empty"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty:/var/empty", "output2:{output}");

    // Step 3: Unset with -evar
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}"))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output3:{output}");

    // Step 4: Pass-through with -evar=
    let output = syd()
        .arg("-plib")
        .arg(&format!("-e{ENV}="))
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output4:{output}");

    // Step 5: Allow with -m trace/allow_unsafe_env:1
    let output = syd()
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_env:1")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output:{output}");

    // Step 6: Toggle -m trace/allow_unsafe_env
    let output = syd()
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_env:1")
        .arg("-mtrace/allow_unsafe_env:0")
        .arg("--")
        .args(["/bin/sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output1:{output}");

    env::remove_var(ENV);

    Ok(())
}

// Tests if `lock:on` command disables access to `/dev/syd`.
fn test_syd_lock() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ bash -c \"test -e /dev/syd\"");
    let status = syd()
        .arg("-plib")
        .args(["--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    eprintln!("+ bash -c \"test -e /dev/syd\"");
    let status = syd()
        .arg("-plib")
        .args(["-mlock:on", "--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Tests if `lock:exec` locks the sandbox for all except the exec child.
fn test_syd_lock_exec() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .arg("-plib")
        .args(["-mlock:exec", "--", "/bin/bash", "-c", "test -e /dev/syd"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .arg("-plib")
        .args([
            "-mlock:exec",
            "--",
            "/bin/bash",
            "-c",
            "( test -e /dev/syd )",
        ])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Tests if syd-chk works to check if process is running under syd.
fn test_syd_chk() -> TestResult {
    eprintln!("+ syd-chk");
    let status = std::process::Command::new(&*SYD_CHK)
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    let status = syd()
        .arg("-plib")
        .args(["--", &SYD_CHK])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_shellexpand_01() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env("SYD_TEST_OOPS", "/home")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_shellexpand_02() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_shellexpand_03() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS:-/home}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_shellexpand_04() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let conf = "allow/write+${SYD_TEST_OOPS:-}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .arg("-plib")
        .arg("-P./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if personality(2) locking works with trace/lock_personality.
fn test_syd_lock_personality() -> TestResult {
    // Check if the target architecture is 32-bit and exit if true
    #[cfg(not(target_pointer_width = "64"))]
    {
        eprintln!("Test requires 64-bit target, skippping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:1")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:0")
        .arg("-mtrace/lock_personality:1")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    // This and the rest of the asserts fail on GitLab CI.
    if !*GL_BUILD {
        assert!(status.success(), "status:{status:?}");
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:0")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "personality")
        .arg("-plib")
        .arg("-mtrace/lock_personality:1")
        .arg("-mtrace/lock_personality:0")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if MDWE protections work as expected.
fn test_syd_mdwe_protections() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:0")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .arg("-mtrace/allow_unsafe_memory:0")
        .args(["--", &SYD_DO, "false"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    let status = syd()
        .env("SYD_DO", "mdwe")
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .arg("-mtrace/allow_unsafe_memory:0")
        .arg("-mtrace/allow_unsafe_memory:1")
        .args(["--", &SYD_DO, "true"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Test if MDWE can be relaxed as expected.
fn test_syd_mdwe_relax() -> TestResult {
    if !is_program_available("luajit") {
        eprintln!("Test requires luajit, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Execute with default restrictions.
    // Expect LuaJIT to segfault.
    let status = syd()
        .arg("-plib")
        .args(["--", "luajit", "-e", "for i=1,1e5 do local a=i*2 end"])
        .status()
        .expect("execute syd");
    // This segfaults on x86-64, but exits with 1 on arm64.
    assert!(
        matches!(status.code().unwrap_or(127), 1 | EX_SIGSEGV),
        "status:{status:?}"
    );

    // Relax restrictions.
    // Expect LuaJIT to succeed.
    let status = syd()
        .arg("-plib")
        .arg("-mtrace/allow_unsafe_memory:1")
        .args(["--", "luajit", "-e", "for i=1,1e5 do local a=i*2 end"])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Test if mmap(NULL, MMAP_FIXED) is prevented.
fn test_syd_mmap_fixed_null() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mmap_fixed_null")
        .arg("-plib")
        .args(["--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_load_library() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    // Write code.
    let code = r#"
#include <stdio.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    assert!(
        matches!(status.code().unwrap_or(127), 42 | nix::libc::ENOSYS),
        "status:{status:?}"
    );

    // Try to load nonexisting library.
    let status = syd()
        .args(["-plib", "--", "./nolib.so"])
        .status()
        .expect("execute syd");
    assert!(
        matches!(
            status.code().unwrap_or(127),
            nix::libc::ENOENT | nix::libc::ENOSYS
        ),
        "status:{status:?}"
    );

    // Try to load a library without the "exec symbol.
    let code = r"int oops(void) { return 42; }";
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    assert!(
        matches!(status.code().unwrap_or(127), 128 | nix::libc::ENOSYS),
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_load_library_noexec() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    // Check for namespaces.
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "-mbind+/:/:noexec",
            "--",
            "./load.so",
        ])
        .status()
        .expect("execute syd");
    assert!(
        matches!(status.code().unwrap_or(127), 42 | nix::libc::ENOSYS),
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_load_library_abort_after_load() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    abort();
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    const XABORT: i32 = 128 + nix::libc::SIGABRT;
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    assert!(
        matches!(status.code().unwrap_or(127), XABORT | nix::libc::ENOSYS),
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_load_library_abort_at_startup() -> TestResult {
    if !is_program_available("cc") {
        eprintln!("Test requires cc, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
    abort(); // This takes down syd with it...
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    let mut file = File::create("load.c")?;
    write!(file, "{code}")?;

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .args(["-plib", "--", "./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code();
    let sign = status.signal();
    assert!(
        sign.unwrap_or(0) == nix::libc::SIGABRT || code.unwrap_or(127) == nix::libc::ENOSYS,
        "status:{status:?}"
    );

    Ok(())
}

// Tests if read sandboxing for open works to allow.
fn test_syd_read_sandbox_open_allow() -> TestResult {
    if !is_program_available("dd") {
        eprintln!("Test requires dd, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ dd if=/dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mallow/read+/***",
            "-mdeny/read+/dev/***",
            "-mallow/read+/dev/null",
            "--",
            "dd",
            "if=/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if read sandboxing for open works to deny.
fn test_syd_read_sandbox_open_deny() -> TestResult {
    if !is_program_available("cat") {
        eprintln!("Test requires cat, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "-mdeny/read+/dev/null",
            "--",
            "cat",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");
    Ok(())
}

// Tests if stat sandboxing for chdir works to allow.
fn test_syd_stat_sandbox_chdir_allow() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdeny/stat+/dev",
            "-mallow/stat+/dev",
            "--",
            "/bin/sh",
            "-c",
            "cd /dev",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_chdir_hide() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ cd /dev");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/dev",
            "--",
            "/bin/sh",
            "-c",
            "cd /dev || exit 42",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 42);
    Ok(())
}

// Tests if stat sandboxing for stat works to allow.
fn test_syd_stat_sandbox_stat_allow() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdeny/stat+/dev/null",
            "-mallow/stat+/dev/null",
            "--",
            "ls",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_stat_hide() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev/null");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/dev/null",
            "--",
            "ls",
            "/dev/null",
        ])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Tests if stat sandboxing for getdents works to allow.
fn test_syd_stat_sandbox_getdents_allow() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev");
    let output = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mdeny/stat+/dev/null",
            "-mallow/stat+/dev/null",
            "--",
            "ls",
            "/dev",
        ])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window == b"null"),
        "Stdout:\n{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing for getdents works to hide.
fn test_syd_stat_sandbox_getdents_hide() -> TestResult {
    if !is_program_available("ls") {
        eprintln!("Test requires ls, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ ls /dev");
    let output = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/dev/null",
            "--",
            "ls",
            "/dev",
        ])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window != b"null"),
        "Stdout:{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing can be bypassed by read attempt
fn test_syd_stat_bypass_with_read() -> TestResult {
    let status = syd()
        .env("SYD_DO", "stat_bypass_with_read")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mdeny/read+/etc/***",
            "-mdeny/stat+/etc/***",
            "-mallow/read+/etc/ld*/***",
            "-mallow/stat+/etc/ld*/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if stat sandboxing can be bypassed by write attempt
fn test_syd_stat_bypass_with_write() -> TestResult {
    let status = syd()
        .env("SYD_DO", "stat_bypass_with_write")
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/stat+/etc/***",
            "-mdeny/write+/etc/***",
            "-mallow/stat+/etc/ld*/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if stat sandboxing can be bypassed by exec attempt
fn test_syd_stat_bypass_with_exec() -> TestResult {
    let status = syd()
        .env("SYD_DO", "stat_bypass_with_exec")
        .args([
            "-ppaludis",
            "-msandbox/exec:on",
            "-msandbox/stat:on",
            "-mallow/exec+/***",
            "-mallow/stat+/***",
            "-mdeny/exec+/**/z?sh",
            "-mdeny/stat+/**/z?sh",
            "-mdeny/exec+/**/[bd]ash",
            "-mdeny/stat+/**/[bd]ash",
            "-mdeny/exec+/**/busybox",
            "-mdeny/stat+/**/busybox",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if write sandboxing for open works to allow.
fn test_syd_write_sandbox_open_allow() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ bash -c \"echo welcome to the machine >> /dev/null\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mdeny/write+/dev/***",
            "-mallow/write+/dev/null",
            "--",
            "/bin/bash",
            "-c",
            "echo welcome to the machine >> /dev/null",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if write sandboxing for open works to deny.
fn test_syd_write_sandbox_open_deny() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ bash -c \"echo welcome to the machine >> /dev/null\"");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "-mdeny/write+/dev/null",
            "--",
            "/bin/bash",
            "-c",
            "echo welcome to the machine >> /dev/null",
        ])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Tests if exec sandboxing works to allow.
fn test_syd_exec_sandbox_open_allow() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let bin = which("true")?;
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-mdeny/exec+/***",
            &format!("-mallow/exec+{bin}"),
            "-atrue", // this may be busybox
            "--",
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

// Tests if exec sandboxing works to deny.
fn test_syd_exec_sandbox_open_deny() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let bin = which("true")?;
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            &format!("-mdeny/exec+{bin}"),
            "-atrue", // this may be busybox
            "--",
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert!(!status.success(), "status:{status:?}");

    Ok(())
}

// Check if shebang program of scripts are properly sandboxed.
fn test_syd_exec_sandbox_deny_shebang() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires cat, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Write script.
    let path = "./script.sh";
    let script = r#"#!/bin/sh -ex
exit 42
"#;
    let mut file = File::create(path)?;
    write!(file, "{script}")?;
    drop(file); // Close the file to avoid ETXTBUSY.

    // Set permissions to make "script.sh" executable.
    let metadata = metadata(path).expect("Failed to read file metadata");
    let mut permissions = metadata.permissions();
    permissions.set_mode(0o755); // This sets the file as executable (rwxr-xr-x).
    set_permissions(path, permissions).expect("Failed to set file permissions");

    // Step 1: Allow both the shebang program and the script.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "--",
            "./script.sh",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 42, "status:{status:?}");

    // Step 2: Allow the shebang program but disable the script.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/***",
            "-mdeny/exec+/**/script.sh",
            "--",
            "./script.sh",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(
        status.code().unwrap_or(127),
        nix::libc::EACCES,
        "status:{status:?}"
    );

    // Step 3: Allow the script but disable the shebang program.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mdeny/exec+/***",
            "-mallow/exec+/**/script.sh",
            "--",
            "./script.sh",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(
        status.code().unwrap_or(127),
        nix::libc::EACCES,
        "status:{status:?}"
    );

    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv4_allow() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+127.0.0.1!4242",
            "-mallow/net/connect+127.0.0.1!4242",
            "--",
            &SYD_DO,
            "127.0.0.1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv4_deny() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+127.0.0.1!4242",
            "-mdeny/net/connect+127.0.0.1!4242",
            "--",
            &SYD_DO,
            "127.0.0.1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 111 /* ECONNREFUSED */);
    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv6_allow() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+::1!4242",
            "-mallow/net/connect+::1!4242",
            "--",
            &SYD_DO,
            "::1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv6_deny() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+::1!4242",
            "-mdeny/net/connect+::1!4242",
            "--",
            &SYD_DO,
            "::1",
            "4242",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 111 /* ECONNREFUSED */);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv4_failure() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+127.0.0.1!0",
            "--",
            &SYD_DO,
            "127.0.0.1",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 111 /* ECONNREFUSED */);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv4_success() -> TestResult {
    let status = syd()
        .env("SYD_DO", "connect4_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:1",
            "-mallow/net/bind+127.0.0.1!0",
            "--",
            &SYD_DO,
            "127.0.0.1",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv6_failure() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:0",
            "-mallow/net/bind+::1!0",
            "--",
            &SYD_DO,
            "::1",
        ])
        .status()
        .expect("execute syd");
    assert_eq!(status.code().unwrap_or(127), 111 /* ECONNREFUSED */);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv6_success() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "connect6_0")
        .args([
            "-mallow/exec,read,stat+/***",
            "-mtrace/allow_safe_bind:1",
            "-mallow/net/bind+::1!0",
            "--",
            &SYD_DO,
            "::1",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_handle_toolong_unix_connect() -> TestResult {
    let status = syd()
        .env("SYD_DO", "toolong_unix_connect")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_sendmsg_scm_rights_one() -> TestResult {
    let status = syd()
        .env("SYD_DO", "sendmsg_scm_rights_one")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_sendmsg_scm_rights_many() -> TestResult {
    let status = syd()
        .env("SYD_DO", "sendmsg_scm_rights_many")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_sendmmsg() -> TestResult {
    let status = syd()
        .env("SYD_DO", "sendmmsg")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_handle_toolong_unix_sendto() -> TestResult {
    let status = syd()
        .env("SYD_DO", "toolong_unix_sendto")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_handle_toolong_unix_sendmsg() -> TestResult {
    let status = syd()
        .env("SYD_DO", "toolong_unix_sendmsg")
        .args([
            "-mallow/exec,read,stat,write+/***",
            "-mallow/net/bind+/***",
            "-mtrace/allow_safe_bind:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exit_wait_all() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "3s");
    let status = syd()
        .args([
            "-ppaludis",
            "-mtrace/exit_wait_all:1",
            "--",
            "/bin/sh",
            "-x",
            "-c",
            "nohup sleep 5 & exit 1",
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    let signal = status.signal().unwrap_or(0);
    let excode = status.code().unwrap_or(127);
    assert!(
        signal == nix::libc::SIGKILL || excode == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_exit_wait_pid() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "3s");
    let status = syd()
        .args([
            "-ppaludis",
            "-mtrace/exit_wait_all:0",
            "--",
            "/bin/sh",
            "-x",
            "-c",
            "nohup sleep 5 & exit 7",
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_eq!(status.code().unwrap_or(127), 7, "status:{status:?}");
    Ok(())
}

fn test_syd_cli_args_override_user_profile() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let _ = unlink(".user.syd-3");
    let mut file = File::create(".user.syd-3").expect("Failed to create .user.syd-3");
    file.write_all(b"mem/max:4242\npid/max:2525\n")
        .expect("Failed to write to .user.syd-3");

    let mut child = syd()
        .args(["-mpid/max:4242", "-mstat", "-ctrue"])
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");

    // Read the output from the child process
    let child_stderr = child.stderr.as_mut().expect("child stderr");
    let mut output = String::new();
    let raw_fd = child_stderr.as_raw_fd();
    let mut file = unsafe { File::from_raw_fd(raw_fd) };
    if let Err(error) = file.read_to_string(&mut output) {
        return Err(TestError(format!(
            "Failed to read output of child process: {error}"
        )));
    }
    print!("Child output:\n{output}");

    assert!(output.contains("Pid Max: 4242"));
    //This may fail if the site-wide config file has lock:on.
    //assert!(output.contains("Memory Max: 4242"));

    Ok(())
}

fn test_syd_force_sandbox() -> TestResult {
    if !is_program_available("true") {
        eprintln!("Test requires true, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let bad_md5 = "a".repeat(32);
    let bad_sha1 = "a".repeat(40);
    let bad_sha256 = "a".repeat(64);
    let bad_sha384 = "a".repeat(96);
    let bad_sha512 = "a".repeat(128);

    // Note, `which" returns canonicalized path.
    let bin_true = which("true").expect("syd-test-do in PATH");

    let md5_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashFunc::Md5,
    )?);
    let sha1_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashFunc::Sha1,
    )?);
    let sha256_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashFunc::Sha256,
    )?);
    let sha384_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashFunc::Sha384,
    )?);
    let sha512_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashFunc::Sha512,
    )?);

    // Test 1: Force sandboxing defaults.
    let status = syd()
        .args(["-plib", "-msandbox/force:on", "--", "true"])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:allow",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:warn",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:deny",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            "-mforce/default:kill",
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );

    // Test 2: Force sandboxing allow with MD5 & SHA1,3-{256,384,512}
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{md5_sum}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{sha1_sum}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{sha256_sum}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{sha384_sum}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{sha512_sum}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Test 3: Force sandboxing deny with hash mismatch.
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_md5}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha1}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha256}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha384}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha512}:deny"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha1}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha256}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha384}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );
    let status = syd()
        .args([
            "-plib",
            "-msandbox/force:on",
            &format!("-mforce+{bin_true}:{bad_sha512}:kill"),
            "--",
            "true",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_exp_symlink_toctou() -> TestResult {
    let status = syd()
        .env("SYD_DO", "symlink_toctou")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_symlinkat_toctou() -> TestResult {
    let status = syd()
        .env("SYD_DO", "symlinkat_toctou")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_ptrmod_toctou_chdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_chdir")
        .args([
            "-plib",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "-mdeny/stat+/var/empty/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    // Note for some reason this test reliable passes on Ubuntu CI,
    // so we do not use fixup! here.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_ptrmod_toctou_exec_fail() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_exec_fail")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/**/toctou_exec",
            "-mallow/exec+/**/syd-test-do*",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    fixup!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_ptrmod_toctou_exec_success() -> TestResult {
    let status = syd()
        .env("SYD_EXEC", "YesPlease")
        .env("SYD_DO", "ptrmod_toctou_exec_success")
        .args([
            "-plib",
            "-msandbox/exec:on",
            "-mallow/exec+/**/toctou_exec",
            "-mallow/exec+/**/syd-test-do*",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    if *CI_BUILD {
        ignore!(status.success(), "status:{status:?}");
    } else {
        fixup!(status.success(), "status:{status:?}");
    }
    Ok(())
}

fn test_syd_exp_ptrmod_toctou_open() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_open")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_ptrmod_toctou_creat() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_creat")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/write+/**/deny.syd-tmp",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_ptrmod_toctou_opath() -> TestResult {
    let status = syd()
        .env("SYD_DO", "ptrmod_toctou_opath")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/etc",
            "-mdeny/stat+/etc/**",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read,stat,write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_io_uring_escape_strict() -> TestResult {
    #[cfg(feature = "uring")]
    {
        // Step 1: Default is strict.
        let status = syd()
            .env("SYD_DO", "io_uring_escape")
            .args([
                "-ppaludis",
                "-msandbox/lock:off",
                "-msandbox/read:on",
                "-msandbox/stat:on",
                "-msandbox/write:on",
                "-mallow/read+/***",
                "-mdeny/stat+/etc/***",
                "-mallow/stat+/etc/ld*",
                "-mdeny/read,write+/etc/passwd",
                "--",
                &SYD_DO,
                "0",
            ])
            .status()
            .expect("execute syd");
        assert!(status.success(), "status:{status:?}");
    }
    Ok(())
}

fn test_syd_io_uring_escape_unsafe() -> TestResult {
    #[cfg(feature = "uring")]
    {
        // Step 2: Relax uring restriction.
        let status = syd()
            .env("SYD_DO", "io_uring_escape")
            .args([
                "-ppaludis",
                "-msandbox/lock:off",
                "-msandbox/read:on",
                "-msandbox/stat:on",
                "-msandbox/write:on",
                "-mallow/read+/***",
                "-mdeny/stat+/etc/***",
                "-mallow/stat+/etc/ld*",
                "-mdeny/read,write+/etc/passwd",
                "-mtrace/allow_unsafe_uring:1",
                "--",
                &SYD_DO,
                "1",
            ])
            .status()
            .expect("execute syd");
        assert!(status.success(), "status:{status:?}");
    }

    Ok(())
}

fn test_syd_opath_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "opath_escape")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mdeny/stat+/etc/***",
            "-mallow/stat+/etc/ld*",
            "-mdeny/read+/etc/passwd",
            "-mdeny/write+/etc/passwd",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_8() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_8")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_9() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_9")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_10() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_10")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_11() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_11")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_12() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_12")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_13() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_13")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_14() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_14")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_15() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_15")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_16() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_16")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_17() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_17")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_18() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_18")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_19() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_19")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_20() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_chdir_relpath_20")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_8() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_8")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_9() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_9")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_10() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_10")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_11() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_11")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_12() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_12")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_13() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_13")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_14() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_14")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_15() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_15")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_16() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_16")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_17() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_17")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_18() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_18")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_19() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_19")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_20() -> TestResult {
    let status = syd()
        .env("SYD_DO", "devfd_escape_open_relpath_20")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_chdir_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_1")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_2")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_3() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_3")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_4")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_5() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_5")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_6() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_6")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_open_relpath_7() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_open_relpath_7")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_relpath() -> TestResult {
    if !is_program_available("grep") {
        eprintln!("Test requires grep, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    eprintln!("+ grep 'Name:[[:space:]]syd' /proc/./self/status");
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            "grep",
            "Name:[[:space:]]syd",
            "/proc/./self/status",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 1,
        "code:{:?}",
        status.code()
    );
    Ok(())
}

fn test_syd_procself_escape_symlink() -> TestResult {
    let status = syd()
        .env("SYD_DO", "procself_escape_symlink")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_procself_escape_symlink_within_container() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .env("SYD_DO", "procself_escape_symlink")
        .args([
            "-mallow/read+/***",
            "-ppaludis",
            "-pcontainer",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_umask_bypass_077() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .env("SYD_DO", "umask_bypass_077")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_umask_bypass_277() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .env("SYD_DO", "umask_bypass_277")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_emulate_opath() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_opath")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-mallow/read+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_emulate_otmpfile() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_otmpfile")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_honor_umask() -> TestResult {
    let prev_umask = umask(Mode::from_bits_truncate(0o077));
    let status = syd()
        .env("SYD_DO", "honor_umask")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
            "0600",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert!(status.success(), "status:{status:?}");

    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .env("SYD_DO", "honor_umask")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
            "0644",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert!(status.success(), "status:{status:?}");

    let prev_umask = umask(Mode::from_bits_truncate(0));
    let status = syd()
        .env("SYD_DO", "honor_umask")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
            "0666",
        ])
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_open_utf8_invalid() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_utf8_invalid")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/write:on",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exec_in_inaccessible_directory() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "exec_in_inaccessible_directory")
        .args([
            "-ppaludis",
            "-msandbox/lock:off",
            "-msandbox/exec:on",
            "-msandbox/write:on",
            "-mallow/exec+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_pipe() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_pipe")
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_socket() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_socket")
        .args([
            "-ppaludis",
            "-msandbox/stat:on",
            "-mallow/stat+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_deleted_file() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_deleted_file")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fstat_on_temp_file() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstat_on_temp_file")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fchmodat_on_proc_fd() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fchmodat_on_proc_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_linkat_on_fd() -> TestResult {
    let status = syd()
        .env("SYD_DO", "linkat_on_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_block_ioctl_tiocsti() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_ioctl_tiocsti")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_block_prctl_ptrace() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_prctl_ptrace")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_block_dev_random() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_dev_random")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "-mdeny/read+/dev/random",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_block_dev_urandom() -> TestResult {
    let status = syd()
        .env("SYD_DO", "block_dev_urandom")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "-mdeny/read+/dev/urandom",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_kill_during_syscall() -> TestResult {
    let status = syd()
        .env("SYD_DO", "kill_during_syscall")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_open_toolong_path() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_toolong_path")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_open_null_path() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_null_path")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_utimensat_null() -> TestResult {
    let status = syd()
        .env("SYD_DO", "utimensat_null")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    // Fails on GitLab CI.
    if !*GL_BUILD {
        assert!(status.success(), "status:{status:?}");
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_normalize_path() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    const NORMALIZE_PATH_TESTS: &[&str] = &[
        "null",
        "./null",
        ".////null",
        ".///.////.///./null",
        "./././././././null",
        "./././.././././dev/null",
        "../dev/././../dev/././null",
    ];

    for path in NORMALIZE_PATH_TESTS {
        let status = syd()
            .args([
                "-ppaludis",
                "-msandbox/read:off",
                "-msandbox/stat:off",
                "-msandbox/write:on",
                "-msandbox/exec:off",
                "-mdeny/write+/***",
                "-mallow/write+/dev/null",
                "--",
                "sh",
                "-c",
                &format!("cd /dev; :> {path}"),
            ])
            .status()
            .expect("execute syd");
        assert_eq!(
            status.code().unwrap_or(127),
            0,
            "path:{path}, status:{status:?}"
        );
    }

    Ok(())
}

fn test_syd_path_resolution() -> TestResult {
    let cwd = syd::fs::canonicalize(
        Pid::this(),
        "/proc/self/cwd",
        None,
        true,
        false,
        syd::fs::MissingHandling::Normal,
    )?
    .to_string_lossy()
    .to_string();

    let status = syd()
        .env("SYD_DO", "path_resolution")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/",
            "-mallow/read+/etc/***",
            "-mallow/read+/lib/***",
            "-mallow/read+/proc/***",
            "-mallow/read+/usr/***",
            "-mallow/stat+/etc/***",
            "-mallow/stat+/lib/***",
            "-mallow/stat+/proc/***",
            "-mallow/stat+/usr/***",
            &format!("-mallow/read+{cwd}/test_file.txt"),
            &format!("-mallow/stat+{cwd}/test_file.txt"),
            &format!("-mallow/write+{cwd}/test_file.txt"),
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_symlink_readonly_path() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/write+/",
            "--",
            "sh",
            "-c",
            "ln -s / test_syd_symlink_readonly_path && unlink test_syd_symlink_readonly_path",
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_remove_empty_path() -> TestResult {
    if !is_program_available("sh") {
        eprintln!("Test requires sh, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            "sh",
            "-c",
            "env LC_ALL=C LANG=C LANGUAGE=C rm '' 2>&1 | tee /dev/stderr | grep -qi 'No such file or directory'"
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_open_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_openat_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "openat_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_lstat_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "lstat_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fstatat_trailing_slash() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fstatat_trailing_slash")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mkdir_trailing_dot() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdir_trailing_dot")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mkdirat_trailing_dot() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdirat_trailing_dot")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_rmdir_trailing_slashdot() -> TestResult {
    let status = syd()
        .env("SYD_DO", "rmdir_trailing_slashdot")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mkdir_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdir_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mkdirat_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdirat_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mknod_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mknod_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mknodat_eexist_escape() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mknodat_eexist_escape")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mdeny/read,stat,write+/boot/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fopen_supports_mode_x() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fopen_supports_mode_x")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fopen_supports_mode_e() -> TestResult {
    let status = syd()
        .env("SYD_DO", "fopen_supports_mode_e")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_link_no_symlink_deref() -> TestResult {
    let status = syd()
        .env("SYD_DO", "link_no_symlink_deref")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_link_posix() -> TestResult {
    let status = syd()
        .env("SYD_DO", "link_posix")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_linkat_posix() -> TestResult {
    let status = syd()
        .env("SYD_DO", "linkat_posix")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_getcwd_long() -> TestResult {
    let status = syd()
        .env("SYD_DO", "getcwd_long")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");

    // FIXME: Fails on musl, possibly not something to fix.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_creat_thru_dangling() -> TestResult {
    let status = syd()
        .env("SYD_DO", "creat_thru_dangling")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_mkdirat_non_dir_fd() -> TestResult {
    let status = syd()
        .env("SYD_DO", "mkdirat_non_dir_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_blocking_udp4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "blocking_udp4")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mtrace/allow_safe_bind:0",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mallow/net/bind+loopback!65432",
            "-mallow/net/connect+loopback!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_blocking_udp6() -> TestResult {
    // Gitlab CI uses docker which has no IPv6.
    if *GL_BUILD {
        eprintln!("IPv6 not available on CI!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .env("SYD_DO", "blocking_udp6")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mtrace/allow_safe_bind:0",
            "-mallow/read,stat,write+/***",
            "-mallow/net/bind+loopback6!65432",
            "-mallow/net/connect+loopback6!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_close_on_exec() -> TestResult {
    let status = syd()
        .env("SYD_DO", "close_on_exec")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_open_exclusive_restart() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_exclusive_restart")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Ignore this on CI!
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    if *CI_BUILD {
        ignore!(status.success(), "status:{status:?}");
    } else {
        assert!(status.success(), "status:{status:?}");
    }
    Ok(())
}

fn test_syd_open_exclusive_repeat() -> TestResult {
    let status = syd()
        .env("SYD_DO", "open_exclusive_repeat")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_pty_io_rust() -> TestResult {
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .env("SYD_DO", "pty_io_rust")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read,stat,write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_pty_io_gawk() -> TestResult {
    if !is_program_available("gawk") {
        eprintln!("Test requires gawk, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .env("SYD_DO", "pty_io_gawk")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_diff_dev_fd() -> TestResult {
    if !is_program_available("diff") {
        eprintln!("Test requires diff, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .env("SYD_DO", "diff_dev_fd")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_fifo_multiple_readers() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            "bash",
            "-c",
        ])
        .arg(
            r#"
# Attempt to DOS syd by spawning multiple FIFO readers in the background.
set -ex
fifo=$(env TMPDIR=. mktemp -u)
mkfifo "$fifo"
pids=()
for i in {1..10}; do
    cat "$fifo" &
    pids+=( $! )
done
# Execute a system call that syd must intervene, this must not block.
touch "$fifo".done
rm -f "$fifo".done
# All good, kill all the cats, wait and exit cleanly.
kill "${pids[@]}" || true
wait
rm -f "$fifo"
# Give syd::m☮☮n thread a chance to clean up for statistics.
sleep 9
true
"#,
        )
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_bind_unix_socket() -> TestResult {
    let status = syd()
        .env("SYD_DO", "bind_unix_socket")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mallow/net/bind+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_signal_protection() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-mdeny/read-/proc/1/**",
            "-mdeny/stat-/proc/1/**",
            "-mdeny/write-/proc/1/***",
            "--",
            "bash",
            "-c",
        ])
        .arg(
            r#"
set -ex
pid=( $(pgrep -w syd) )
# Sending signal 0 must work.
for tid in "${pid[@]}"; do
    kill -0 ${tid}
    sleep 1
done
# Sending other signals are not permitted.
for sig in INT ABRT STOP KILL; do
    for tid in "${pid[@]}"; do
        kill -${sig} ${tid} && exit 1
        sleep 1
    done
done
"#,
        )
        .status()
        .expect("execute syd");

    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_emulate_open_fifo_1() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_open_fifo_1")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_emulate_open_fifo_2() -> TestResult {
    let status = syd()
        .env("SYD_DO", "emulate_open_fifo_2")
        .args([
            "-ppaludis",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_deny_magiclinks() -> TestResult {
    // Test expects PID namespace.
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    // Check protections with stat sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "deny_magiclinks")
        .args([
            "-ppaludis",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Check protections with stat sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "deny_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Check protections with stat sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "deny_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "-mlock:on",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_lstat_magiclinks() -> TestResult {
    // Test expects PID namespace.
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    // Check protections with stat sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "lstat_magiclinks")
        .args([
            "-ppaludis",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Check protections with stat sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "lstat_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Check protections with stat sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "lstat_magiclinks")
        .args([
            "-plib",
            "-munshare/user:1",
            "-munshare/pid:1",
            "-mlock:on",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_access_unsafe_paths() -> TestResult {
    // Check protections with read, stat, write sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "access_unsafe_paths")
        .args(["-ppaludis", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Check protections with sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "access_unsafe_paths")
        .args(["-plib", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    // Check protections with sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .env("SYD_DO", "access_unsafe_paths")
        .args(["-plib", "-mlock:on", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_unshare_user_bypass_limit() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    let status = syd()
        .env("SYD_DO", "unshare_user_bypass_limit")
        .args(["-plib", "-pcontainer", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_exp_interrupt_mkdir() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_mkdir")
        .args([
            "-ppaludis",
            "-mtrace/allow_safe_bind:f",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Ignore this on CI!
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    if *CI_BUILD {
        ignore!(status.success(), "status:{status:?}");
    } else {
        assert!(status.success(), "status:{status:?}");
    }
    Ok(())
}

fn test_syd_exp_interrupt_bind_ipv4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_bind_ipv4")
        .args([
            "-ppaludis",
            "-mtrace/allow_safe_bind:f",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Ignore this on CI!
    // FIXME: This is a kernel bug, mixi will report it, they have a POC.
    if *CI_BUILD {
        ignore!(status.success(), "status:{status:?}");
    } else {
        assert!(status.success(), "status:{status:?}");
    }
    Ok(())
}

fn test_syd_exp_interrupt_bind_unix() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_bind_unix")
        .args([
            "-ppaludis",
            "-mtrace/allow_safe_bind:f",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mallow/net/bind+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Ignore this on CI!
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    if *CI_BUILD {
        ignore!(status.success(), "status:{status:?}");
    } else {
        assert!(status.success(), "status:{status:?}");
    }
    Ok(())
}

fn test_syd_exp_interrupt_connect_ipv4() -> TestResult {
    let status = syd()
        .env("SYD_DO", "interrupt_connect_ipv4")
        .args([
            "-ppaludis",
            "-mtrace/allow_safe_bind:f",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/write+/***",
            "-mallow/net/bind+loopback!65432",
            "-mallow/net/connect+loopback!65432",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Ignore this on CI!
    // FIXME: This is a kernel bug, mixi will report it, they have a POC.
    if *CI_BUILD {
        ignore!(status.success(), "status:{status:?}");
    } else {
        assert!(status.success(), "status:{status:?}");
    }
    Ok(())
}

/*
fn test_syd_repetitive_clone() -> TestResult {
    if !is_program_available("bash") {
        eprintln!("Test requires bash, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    let status = syd()
        .env("SYD_DO", "repetitive_clone")
        .args(["-ppaludis", "--", &SYD_DO])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(0) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );
    Ok(())
}
*/

fn test_syd_exp_syscall_fuzz_bare() -> TestResult {
    let epoch = std::time::Instant::now();
    let status = syd()
        .env("SYD_DO", "syscall_fuzz")
        .args([
            "-ppaludis",
            "-mlock:off",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/exec:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/exec+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_exp_syscall_fuzz_wrap() -> TestResult {
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    }

    let epoch = std::time::Instant::now();
    let status = syd()
        .env("SYD_DO", "syscall_fuzz")
        .args([
            "-ppaludis",
            "-pcontainer",
            "-mlock:off",
            "-msandbox/lock:off",
            "-msandbox/read:on",
            "-msandbox/stat:on",
            "-msandbox/write:on",
            "-msandbox/exec:on",
            "-msandbox/net:on",
            "-mallow/read+/***",
            "-mallow/stat+/***",
            "-mallow/exec+/***",
            "-mallow/write+/***",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_pid_thread_deny() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let status = syd()
        .env("SYD_DO", "thread")
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:1",
            "--",
            &SYD_DO,
            "0",
            "24",
        ])
        .status()
        .expect("execute syd");
    // Early fails get EACCES ie 13.
    // In debug mode, this fails with 101.
    // In release mode, this fails with 134.
    // Both indicate thread::spawn paniced.
    assert!(
        matches!(status.code().unwrap_or(127), 13 | 101 | 134),
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_pid_fork_deny() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let status = syd()
        .env("SYD_DO", "fork")
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
            "0",
            "24",
        ])
        .status()
        .expect("execute syd");
    env::remove_var("SYD_DO");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_pid_fork_bomb() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {
            eprintln!("PID sandboxing works in a container!");
            eprintln!("Proceeding with test...");
        }
    }

    eprintln!("Running the fork bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_DO", "fork_bomb")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-plib",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_pid_fork_bomb_asm() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {
            eprintln!("PID sandboxing works in a container!");
            eprintln!("Proceeding with test...");
        }
    }

    eprintln!("Running the fork bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_DO", "fork_bomb_asm")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-plib",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_pid_thread_bomb() -> TestResult {
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {
            eprintln!("PID sandboxing works in a container!");
            eprintln!("Proceeding with test...");
        }
    }

    eprintln!("Running the thread bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_DO", "thread_bomb")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-plib",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:16",
            "--",
            &SYD_DO,
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_exp_pid_stress_ng_kill() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:1",
            "-mpid/kill:t",
            "--",
            "stress-ng",
            "-c",
            "1",
            "-t",
            "7",
        ])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_exp_pid_stress_ng_allow() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let _ = remove_file("log");
    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:2",
            "--",
            "stress-ng",
            "--log-file",
            "log",
            "-c",
            "1",
            "-t",
            "7",
        ])
        .status()
        .expect("execute syd");
    // Fails on CI.
    if !*CI_BUILD {
        assert!(status.success(), "status:{status:?}");
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let mut file = File::open("log")?;
    let mut logs = String::new();
    file.read_to_string(&mut logs)?;

    assert!(!logs.contains("errno="), "logs:{logs:?}");

    Ok(())
}

fn test_syd_exp_pid_stress_ng_fork() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_pid() {
        Some(false) => {
            eprintln!("Test requires PID sandboxing to work in a container, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError("Failed to test for Pid Sanboxing!".to_string()));
        }
        _ => {}
    }

    let _ = remove_file("log");
    let status = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/pid:on",
            "-mpid/max:128",
            "-mfilter/pid:1",
            "--",
            "stress-ng",
            "--log-file",
            "log",
            "-f",
            "4",
            "-t",
            "15",
            "--fork-max",
            "1024",
        ])
        .status()
        .expect("execute syd");
    // Fails on CI.
    if !*CI_BUILD {
        assert!(status.success(), "status:{status:?}");
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let mut file = File::open("log")?;
    let mut logs = String::new();
    file.read_to_string(&mut logs)?;
    assert!(!logs.contains("errno="), "logs:{logs:?}");

    Ok(())
}

fn test_syd_mem_alloc() -> TestResult {
    let status = syd()
        .env("SYD_DO", "alloc")
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .args([
            "-ppaludis",
            "-msandbox/mem:on",
            "-mmem/max:128M",
            "-mmem/vm_max:256M",
            "--",
            &SYD_DO,
        ])
        .status()
        .expect("execute syd");
    // Segmentation fault is expected.
    // Iot is confusing but happens on alpine+musl.
    // Otherwise we require ENOMEM.
    assert!(
        matches!(
            status.code().unwrap_or(127),
            nix::libc::ENOMEM | EX_SIGIOT | EX_SIGSEGV
        ),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_exp_mem_stress_ng_malloc_1() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let command = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/mem:on",
            "-mmem/max:32M",
            "-mmem/vm_max:256M",
            "--",
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--malloc",
            "4",
            "--malloc-bytes",
            "128M",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    assert!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

fn test_syd_exp_mem_stress_ng_malloc_2() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let command = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/mem:on",
            "-mmem/max:32M",
            "-mmem/vm_max:256M",
            "--",
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--malloc",
            "4",
            "--malloc-bytes",
            "128M",
            "--malloc-touch",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    assert!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

fn test_syd_exp_mem_stress_ng_mmap() -> TestResult {
    if !is_program_available("stress-ng") {
        eprintln!("Test requires stress-ng, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    };
    match check_unshare() {
        Some(false) => {
            eprintln!("Test requires Linux namespaces, skipping!");
            env::set_var("SYD_TEST_SOFT_FAIL", "1");
            return Ok(());
        }
        None => {
            return Err(TestError(
                "Failed to test for Linux namespaces!".to_string(),
            ));
        }
        _ => {}
    };

    let command = syd()
        .args([
            "-ppaludis",
            "-pcontainer",
            "-msandbox/mem:on",
            "-mmem/max:16M",
            "-mmem/vm_max:64M",
            "--",
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--mmap",
            "4",
            "--mmap-bytes",
            "1G",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    fixup!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}
